@@ -6,10 +6,274 @@ import re
 import string
 import random
 import shutil
+import importlib
 from zipfile import ZipFile
 from pathlib import Path
 
 
+##########
+# Aurora #
+##########
+##
+# @desc Generates controller routes
+# 
+# @param controller: str -- The controller class name
+# @param url: str -- The controller url
+# @param methods: list -- The controller REST methods
+# 
+# @return tuple
+##
+def controller(name:str, url:str='', methods:list=['GET']):
+    # Check required params
+    if not name:
+        # Raise error
+        raise Exception("You must provide the required parameters: ['name']")
+
+    # Check controller name
+    if not controller_name(name)['result']:
+        # Raise error
+        raise Exception(controller_name(name)['message'])
+
+    # Check controller url
+    if url and not controller_url(url)['result']:
+        # Raise error
+        raise Exception(controller_url(url)['message'])
+
+    # Check controller methods
+    if  not controller_methods(methods)['result']:
+        # Raise error
+        raise Exception(controller_methods(methods)['message'])
+
+    return (name, url, methods)
+
+
+##
+# @desc Generates app route
+# 
+# @param name: str -- The app name
+# @param url: str -- The app base url
+# 
+# @return tuple
+##
+def app(name:str, url:str):
+    # Check required params
+    if not name or not url:
+        # Raise error
+        raise Exception("You must provide the required parameters: ['name', 'url']")
+
+    # Check app name
+    if not app_name(name)['result']:
+        # Raise error
+        raise Exception(app_name(name)['message'])
+
+    # Check app url
+    if not app_url(url)['result']:
+        # Raise error
+        raise Exception(app_url(url)['message'])
+
+    return (name, url)
+
+
+##
+# @desc Check if an app exists
+# 
+# @param app: str -- The app name
+# @param apps: list -- The apps list
+# 
+# @return dict
+##
+def app_exists(app:str):
+    # Apps info
+    apps_module = importlib.import_module(f'_apps')
+    apps = getattr(apps_module, 'apps')
+
+    exists = False
+    url = ''
+    while True:
+        i = 0
+        for route in apps:
+            # App exists
+            if app == route[0]:
+                url = route[1]
+                exists = True
+                break
+
+            i += 1
+
+        break
+
+    # App exists
+    if exists:
+        return {
+            'result': True, 
+            'url': f'{url}'
+        }
+
+    # App not exists
+    else:
+        return {
+            'result': False, 
+            'message': f'The "{app}" app doesn\'t exist!'
+        }
+
+
+##
+# @desc Check if an app url exists
+# 
+# @param url: str -- The app url
+# 
+# @return bool
+##
+def app_url_exists(url:str):
+    # Apps info
+    apps_module = importlib.import_module(f'_apps')
+    apps = getattr(apps_module, 'apps')
+
+    exists = False
+    while True:
+        i = 0
+        for route in apps:
+            # App url exists
+            if url == route[1]:
+                exists = True
+                break
+
+            i += 1
+
+        break
+
+    # App url exists
+    if exists:
+        return True
+
+    # App url not exists
+    else:
+        return False
+
+
+##
+# @desc Check if an controller exists
+# 
+# @param app: str -- The app name
+# @param controller: str -- The controller name
+# @param controller: list -- The controllers list
+# 
+# @return dict
+##
+def controller_exists(app:str, controller:str):
+    # Controllers info
+    controllers_module = importlib.import_module(f'controllers.{app}._controllers')
+    controllers = getattr(controllers_module, 'controllers')
+
+    exists = False
+    url = ''
+    while True:
+        i = 0
+        for route in controllers:
+            # App exists
+            if controller == route[0]:
+                url = route[1]
+                exists = True
+                break
+
+            i += 1
+
+        break
+
+    # App exists
+    if exists:
+        return {
+            'result': True, 
+            'url': f'{url}'
+        }
+
+    # App not exists
+    else:
+        return {
+            'result': False, 
+            'message': f'The "{app}" app doesn\'t have the "{controller}" controller!'
+        }
+
+
+##
+# @desc Check if an app url exists
+# 
+# @param app: str -- The app name
+# @param url: str -- The app url
+# 
+# @return dict
+##
+def controller_url_exists(app:str, url:str=''):
+    # Controllers info
+    controllers_module = importlib.import_module(f'controllers.{app}._controllers')
+    controllers = getattr(controllers_module, 'controllers')
+
+    exists = False
+    while True:
+        i = 0
+        for route in controllers:
+            # Controller url exists
+            if url == route[1]:
+                exists = True
+                break
+
+            i += 1
+
+        break
+
+    # Controller url exists
+    if exists:
+        return True
+
+    # Controller url not exists
+    else:
+        return False
+
+
+##
+# @desc Produces the final url for a route (app url + controller url)
+# 
+# @param app: str - The app name
+# @param controller: str - The app controller name
+#
+# @return object
+##
+def route_url(app:str, controller:str=None):
+    # Apps info
+    apps_module = importlib.import_module(f'_apps')
+    apps = getattr(apps_module, 'apps')
+
+    app_url = ''
+
+    # App not exists
+    if not app_exists(app)['result']:
+        # Raise error
+        raise Exception(app_exists(app)['message'])
+
+    # App exists
+    else:
+        app_url = app_exists(app)['url']
+
+    # Controller inserted
+    if controller:
+        # Controller not exists
+        if not controller_exists(app, controller)['result']:
+            # Raise error
+            raise Exception(controller_exists(app, controller)['message'])
+
+        # Controller exists
+        else:
+            controller_url = controller_exists(app, controller)['url']
+
+        url = f'/{app_url}/{controller_url}/'
+
+    # Controller not inserted
+    else:
+        url = f'/{app_url}/'
+
+    return url
+
+
 ###################
 # String Handling #
 ###################
@@ -33,7 +297,56 @@ def random_string(size:int=8, chars:str=string.ascii_uppercase + string.digits):
 # @retun str
 ##
 def snake_case(name:str):
-    return re.sub(r'(?<!^)(?=[A-Z])', '_', name).lower()
+    result = ""
+
+    result = re.sub(r'(?<!^)(?=[A-Z])', '_', name).lower()
+
+    return result
+
+
+##
+# @desc Converts string to snake_case
+#
+# @param name: str
+#
+# @retun str
+##
+def to_snake_case(name:str):
+    result = ''
+    underscore = False
+
+    # Remove the spaces
+    name = delete_chars(name, ' ')
+
+    # Loop the name characters
+    for i in range(len(name)):
+        
+        # Check if the next character is underscore
+        if i < len(name) - 1:
+            underscore = name[i + 1] == "_" or name[i + 1].isupper()
+
+        # Ignore duplicated underscores
+        if name[i] == "_" and underscore:
+            continue
+
+        # Accept single underscores
+        elif name[i] == "_":
+            result += "_"
+
+        # Convert uppercase characters
+        elif name[i].isupper():
+            result += "_" + name[i].lower()
+
+        # Accept other characters
+        else:
+            result += name[i]
+
+    # Remove first character underscore
+    if result[0] == "_":
+        result = result[1:]
+
+    # Return converted result
+    return result
 
 
 ##
@@ -71,7 +384,7 @@ def dict_factory(cur:object, row:object):
 
 
 ##
-# @desc Convert named list into dictionary list, for PostgreSQL Database
+# @desc Convert named list into dictionary list, for Postgres Database
 #
 # @param cur: list - a named list
 #
@@ -88,9 +401,56 @@ def real_dict(cur:list):
     return translate
 
 
+##
+# @desc Checks if there are duplicates in a list
+#
+# @param ls: list
+#
+# @retun bool
+##
+def list_dup(ls:list):
+    return len(ls) != len(set(ls))
+
+
+##
+# @desc Checks if there are duplicate values in a dictionary
+#
+# @param dt: dict
+#
+# @retun bool
+##
+def dict_dup_val(dt:dict):
+    new_dt = {}
+
+    for key, value in dt.items():
+        new_dt.setdefault(value, set()).add(key)
+
+    if [key for key, values in new_dt.items() if len(values) > 1]:
+        return True
+    else:
+        return False
+
+
 #################
 # File Handling #
 #################
+##
+# @desc Checks if a file exists
+#
+# @param file: str - The absolute file path
+#
+# @retun bool
+##
+def file_exist(file:str):
+    # File exists
+    if os.path.exists(file):
+        return True
+
+    # File not exists
+    else:
+        return False
+
+
 ##
 # @desc Validates file name
 #
@@ -322,21 +682,29 @@ def unzip_file(file_path:str, dest_dir:str):
 # @param file_path: str - The absolute file path
 # @param old_str: str - The old string
 # @param new_str: str - The new string
+# @param regex: bool - For replacing a regular expression
 #
 # @retun bool -- True (on success), False (on error)
 ##
-def replace_file_string(file_path:str, old_str:str, new_str:str):
+def replace_file_string(file_path:str, old_str:str, new_str:str, regex:bool=False):
     # Read the file
     with open(file_path, 'r') as file :
         f = file.read()
 
-    # Replace the string
-    f = f.replace(old_str, new_str)
+    # Replace regular expression
+    if regex:
+        f = re.sub(old_str, new_str, f, flags = re.M)
+    
+    # Replace string
+    else:
+        f = f.replace(old_str, new_str)
 
     # Write the file out again
     with open(file_path, 'w') as file:
         file.write(f)
 
+    return file.close()
+
 
 ##
 # @desc Replace lines in a file contain a string with new line data
@@ -344,10 +712,11 @@ def replace_file_string(file_path:str, old_str:str, new_str:str):
 # @param file_path: str - The absolute file path
 # @param old_line: str - The character to match in the line
 # @param new_str: str - The new line data
+# @param regex: bool - For replacing a regular expression
 #
 # @retun bool -- True (on success), False (on error)
 ##
-def replace_file_line(file_path:str, old_line:str, new_line:str):
+def replace_file_line(file_path:str, old_line:str, new_line:str, regex:bool=False):
     # Open file
     with open(file_path, 'r+') as f:
         
@@ -357,19 +726,35 @@ def replace_file_line(file_path:str, old_line:str, new_line:str):
         # Set the position to the beginning of the file 
         f.seek(0)
 
-        # Loop the lines
-        for line in lines:
-            # Replace old lines
-            if old_line in line:
-                f.write(new_line)
+        # Replace regular expression
+        if regex:
+            # Loop the lines
+            for line in lines:
+                # Replace old lines
+                if re.match(old_line, line):
+                    f.write(new_line)
 
-            # Write other lines
-            else:
-                f.write(line)
+                # Write other lines
+                else:
+                    f.write(line)
+        
+        # Replace string
+        else:
+            # Loop the lines
+            for line in lines:
+                # Replace old lines
+                if old_line in line:
+                    f.write(new_line)
+
+                # Write other lines
+                else:
+                    f.write(line)
 
         # Resize the file to the current file stream position
         f.truncate()
 
+    return f.close()
+
 
 ######################
 # Directory Handling #
@@ -486,18 +871,18 @@ def app_name(name:str):
 
 
 ##
-# @desc Validates base url
+# @desc Validates app url
 #
-# @param url: str - The app name
+# @param url: str - The app url
 #
 # @retun dict
 ##
-def base_url(url:str):
-    # Check required base URL
+def app_url(url:str):
+    # Check required app URL
     if not url:
         return {
             'result': False, 
-            'message': 'The base URL is required!'
+            'message': 'The app URL is required!'
         }
 
     # Regular expression
@@ -514,7 +899,7 @@ def base_url(url:str):
     else:
         return {
             'result': False, 
-            'message': 'The base URL is invalid.\nValid characters: a-z, -'
+            'message': 'The app URL is invalid.\nValid characters: a-z, -'
         }
 
 
@@ -567,7 +952,7 @@ def controller_url(url:str):
         }
 
     # Regular expression
-    regex = '^[a-z]+[a-z-<>:/]*$'
+    regex = '^[a-z0-9]+[a-z0-9-<>:/]*$'
 
     # Valid URL
     if re.match(regex, url):
@@ -580,7 +965,7 @@ def controller_url(url:str):
     else:
         return {
             'result': False, 
-            'message': 'The controller URL is invalid.\nValid characters: a-z, -, /, <, :, >'
+            'message': 'The controller URL is invalid.\nValid characters: a-z, 0-9, -, /, <, :, >'
         }
 
 
@@ -717,3 +1102,69 @@ def form_name(name:str):
             'message': 'The form name must be in "CamelCase" form with at least two "a-z" and "A-Z" characters.'
         }
 
+
+##
+# @desc Validates database table names
+#
+# @param name: str - The app name
+#
+# @retun dict
+##
+def table_name(name:str):
+    # Check required table name
+    if not name:
+        return {
+            'result': False, 
+            'message': 'The table name is required!'
+        }
+
+    # Regular expression
+    regex = '^[a-z_]+$'
+
+    # Valid name
+    if re.match(regex, name) and name == to_snake_case(name) and len(name) >= 2:
+        return {
+            'result': True, 
+            'message': ''
+        }
+
+    # Invalid name
+    else:
+        return {
+            'result': False, 
+            'message': 'Database table names must be in "snake_case" form with at least two a-z, _ characters.'
+        }
+
+
+##
+# @desc Validates database column names
+#
+# @param name: str - The app name
+#
+# @retun dict
+##
+def column_name(name:str):
+    # Check required column name
+    if not name:
+        return {
+            'result': False, 
+            'message': 'The column name is required!'
+        }
+
+    # Regular expression
+    regex = '^[a-z_]+$'
+
+    # Valid name
+    if re.match(regex, name) and name == to_snake_case(name) and len(name) >= 2:
+        return {
+            'result': True, 
+            'message': ''
+        }
+
+    # Invalid name
+    else:
+        return {
+            'result': False, 
+            'message': 'Database column names must be in "snake_case" form with at least two a-z, _ characters.'
+        }
+
